"""Global variables associated to the test suite."""

#: List of CPP variables that should be defined in config.h in order to enable this suite.
need_cpp_vars = [
]

#: List of keywords that are automatically added to all the tests of this suite. 
keywords = [
]

#: List of input files
inp_files = [
"t01.in",
"t02.in",
"t03.in",
"t04.in",
"t05.in",
"t06.in",
"t07.in",
"t08.in",
"t09.in",
"t17.in",
"t20.in",
"t21.in",
"t22.in",
"t23.in",
"t24.in",
"t25.in",
"t26.in",
"t27.in",
"t28.in",
"t29.in",
"t30.in",
"t31.in",
"t32.in",    
"t33.in",
"t34.in",
"t35.in",
"t36.in",
"-t37.in", # Test 37 is not portable, and has been disabled
"t38.in",
"t39.in",
"t40.in",
"t41.in",
"t42.in",
"t43.in",
"t44.in",
"t45.in",
"t46.in",
"t50.in", 
"t51.in", 
"t52.in",
"t53.in",
"t54.in",
"-t55.in", # Disabled!
"t56.in", 
"t57.in", 
"t58.in",
"t59.in",
"t60.in",
"t61.in",
"t62.in",
"t63.in",
"t64.in",
"t65.in",
"t66.in",
"t67.in",
"t68.in",
"t69.in",
"t70.in",
"t71.in",
"t72.in",
"t73.in",
"t74.in",
"t75.in",
"t76.in",
"t77.in",
"-t78.in", # transferred in seq
"t79.in",
"-t80.in", # transferred in seq
"t81.in",
"t82.in",
"t83.in",
"t84.in",
"t85.in",
"t86.in",
"t87.in",
"t88.in",
"t89.in",
"-t90.in", # transferred in seq
"t91.in",
"t92.in",
"t93.in",
"t94.in",
"t95.in",
"t96.in",  
"t97.in",
"t98.in",
"t99.in",
]
