"""Global variables associated to the test suite."""

#: List of CPP variables that should be defined in config.h in order to enable this suite.
need_cpp_vars = [
]

#: List of keywords that are automatically added to all the tests of this suite. 
keywords = [
]

#: List of input files
inp_files = [
"t01.in",
"t02.in",
"t03.in",
"t04.in",
"t05.in",
"t06.in",
"t07.in",
"t08.in",
"t09.in",
"t11.in",
"t12.in",
"t13.in",
"t14.in",
"t15.in",
"t16.in",
"t19.in",
"t21.in",
"t22.in",
"t29.in",
"t31.in",
"t32.in",
"t33.in",
"t34.in",
"t35.in",
"t36.in",
"t37.in",
"t38.in",
"t39.in",
"t40.in",
"t41.in",
"t50.in",
"t51.in",
"t52.in",
"t53.in",
]
