"""Global variables associated to the test suite."""

#: List of CPP variables that should be defined in config.h in order to enable this suite.
need_cpp_vars = [
"HAVE_LIBXC",
]

#: List of keywords that are automatically added to all the tests of this suite. 
keywords = [
]

#: List of input files
inp_files = [
"t00.in",
"t01.in",
"t02.in",
"t03.in",
"t04.in",
"t05.in",
"t06.in",
"t07.in",
"t08.in",
"t09.in",
"t10.in",
"-t11.in", # Disabled
"-t12.in", # Disabled
"t13.in",
"-t14.in", # Disabled
"-t15.in", # Disabled
"-t16.in", # Disabled
"t17.in",
"t18.in",
"t19.in",
"t20.in",
"t21.in",
"t41.in",
"t42.in", 
"t43.in", 
"t44.in", 
"t45.in", 
"t51.in", 
"t52.in", 
"t53.in",  
"t67.in",  
"t68.in",  
"t69.in",  
"t70.in",  
"t71.in",  
"t72.in", 
"t73.in", 
"t74.in", 
"t81.in", 
"t82.in", 
]
