"""Global variables associated to the test suite."""

#: List of CPP variables that should be defined in config.h in order to enable this suite.
need_cpp_vars = [
"HAVE_LIBPSML",
]

#: List of keywords that are automatically added to all the tests of this suite. 
keywords = [
"PSML"
]

#: List of input files
inp_files = [
"t01.in",
"t02.in",
"t03.in",
"t04.in",
"t05.in",
"t06.in",
"t07.in",
"t08.in",
"t09.in",
"t10.in",
"t11.in",
"t12.in",
"t13.in",
"t14.in",
]
